@@ -13,7 +13,7 @@
 #include <afsconfig.h>
 #include "../afs/param.h"
 
-RCSID("$Header: /cvs/openafs/src/afs/afs_segments.c,v 1.5.2.2 2002/06/08 00:31:42 shadow Exp $");
+RCSID("$Header: /cvs/openafs/src/afs/afs_segments.c,v 1.10 2001/12/25 18:10:39 shadow Exp $");
 
 #include "../afs/sysincludes.h" /*Standard vendor system headers*/
 #include "../afs/afsincludes.h" /*AFS-based standard headers*/
@@ -60,7 +60,7 @@ int afs_StoreMini(avc, areq)
     struct AFSVolSync tsync;
     register afs_int32 code;
     register struct rx_call *tcall;
-    afs_int32 tlen;
+    afs_size_t tlen, base = 0;
     XSTATS_DECLS
 
     AFS_STATCNT(afs_StoreMini);
@@ -77,6 +77,7 @@ int afs_StoreMini(avc, areq)
 #ifdef RX_ENABLE_LOCKS
 	    AFS_GUNLOCK();
 #endif /* RX_ENABLE_LOCKS */
+retry:
 	    tcall = rx_NewCall(tc->id);
 #ifdef RX_ENABLE_LOCKS
 	    AFS_GLOCK();
@@ -92,14 +93,42 @@ int afs_StoreMini(avc, areq)
 	    InStatus.Mask = AFS_SETMODTIME;
 	    InStatus.ClientModTime = avc->m.Date;
 	    XSTATS_START_TIME(AFS_STATS_FS_RPCIDX_STOREDATA);
+            afs_Trace4(afs_iclSetp, CM_TRACE_STOREDATA64, 
+		ICL_TYPE_FID, &avc->fid.Fid,
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(base), 
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(tlen), 
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(avc->m.Length));
 #ifdef RX_ENABLE_LOCKS
 	    AFS_GUNLOCK();
 #endif /* RX_ENABLE_LOCKS */
+#ifdef AFS_64BIT_CLIENT
+	    if (!afs_serverHasNo64Bit(tc)) {
+	        code = StartRXAFS_StoreData64(tcall,
+					(struct AFSFid *)&avc->fid.Fid,
+					&InStatus, avc->m.Length,
+					base, tlen);
+	    } else {
+		afs_int32 l1, l2;
+		l1 = avc->m.Length;
+		l2 = tlen;
+	    	code = StartRXAFS_StoreData(tcall,
+					(struct AFSFid *)&avc->fid.Fid,
+					&InStatus, l1, 0, l2);
+	    }
+#else /* AFS_64BIT_CLIENT */
 	    code = StartRXAFS_StoreData(tcall,
 					(struct AFSFid *)&avc->fid.Fid,
 					&InStatus, avc->m.Length, 0, tlen);
+#endif /* AFS_64BIT_CLIENT */
 	    if (code == 0) {
 		code = EndRXAFS_StoreData(tcall, &OutStatus, &tsync);
+#ifdef AFS_64BIT_CLIENT
+	        if (code == RXGEN_OPCODE) {
+		    afs_serverSetNo64Bit(tc);
+	            code = rx_EndCall(tcall, code);
+		    goto retry;
+		}
+#endif /* AFS_64BIT_CLIENT */
 	    }
 	    code = rx_EndCall(tcall, code);
 #ifdef RX_ENABLE_LOCKS
@@ -118,7 +147,7 @@ int afs_StoreMini(avc, areq)
     }
     else {
 	/* blew it away */
-	afs_InvalidateAllSegments(avc, 1);
+	afs_InvalidateAllSegments(avc);
     }
     return code;
 
@@ -161,15 +190,15 @@ afs_StoreAllSegments(avc, areq, sync)
     afs_hyper_t newDV, oldDV;	/* DV when we start, and finish, respectively */
     struct dcache **dcList, **dclist;
     unsigned int i, j, minj, maxj, moredata, high, off;
-    unsigned long tlen;
+    afs_size_t tlen;
+    afs_size_t maxStoredLength; /* highest offset we've written to server. */
     int safety;
-    int maxStoredLength; /* highest offset we've written to server. */
 #ifndef AFS_NOSTATS
     struct afs_stats_xferData *xferP;	/* Ptr to this op's xfer struct */
     osi_timeval_t  xferStartTime,	/*FS xfer start time*/
                    xferStopTime;	/*FS xfer stop time*/
-    afs_int32 bytesToXfer;			/* # bytes to xfer*/
-    afs_int32 bytesXferred;			/* # bytes actually xferred*/
+    afs_size_t bytesToXfer;			/* # bytes to xfer*/
+    afs_size_t bytesXferred;			/* # bytes actually xferred*/
 #endif /* AFS_NOSTATS */
 
 
@@ -181,7 +210,7 @@ afs_StoreAllSegments(avc, areq, sync)
     foreign = (avc->states & CForeign);
     dcList = (struct dcache **) osi_AllocLargeSpace(AFS_LRALLOCSIZ);
     afs_Trace2(afs_iclSetp, CM_TRACE_STOREALL, ICL_TYPE_POINTER, avc,
-	       ICL_TYPE_INT32, avc->m.Length);
+	       ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(avc->m.Length));
 #if !defined(AFS_AIX32_ENV) && !defined(AFS_SGI65_ENV)
     /* In the aix vm implementation we need to do the vm_writep even
      * on the memcache case since that's we adjust the file's size
@@ -250,8 +279,8 @@ afs_StoreAllSegments(avc, areq, sync)
 	if ((afs_indexFlags[index] & IFDataMod) &&
 	    (afs_indexUnique[index] == avc->fid.Fid.Unique)) {
 	  tdc = afs_GetDSlot(index, 0);  /* refcount+1. */
+	  ReleaseReadLock(&tdc->tlock);
 	  if (!FidCmp( &tdc->f.fid, &avc->fid ) && tdc->f.chunk >= minj ) {
-
 	    off = tdc->f.chunk - minj;
 	    if (off < NCHUNKSATONCE) {
 	      if ( dcList[ off ] )
@@ -259,34 +288,39 @@ afs_StoreAllSegments(avc, areq, sync)
 	      dcList[ off ] = tdc;
 	      if (off > high) 
 		high = off;
-	      tlen -= tdc->f.chunkBytes; /* shortcut: big win for little files */
 	      j++;
-	      if (tlen <= 0)
-		break;
+	      /* DCLOCKXXX: chunkBytes is protected by tdc->lock which we
+	       * can't grab here, due to lock ordering with afs_xdcache.
+	       * So, disable this shortcut for now.  -- kolya 2001-10-13
+	       */
+	      /* shortcut: big win for little files */
+	      /* tlen -= tdc->f.chunkBytes;
+	       * if (tlen <= 0)
+	       *    break;
+	       */
 	    }
 	    else {
 	      moredata = TRUE;
-	      lockedPutDCache(tdc);
+	      afs_PutDCache(tdc);
 	      if (j == NCHUNKSATONCE)
 		break;
 	    }
 	  } else {
-	    lockedPutDCache(tdc);
+	    afs_PutDCache(tdc);
 	  }
 	}
 	index = afs_dvnextTbl[index];
       }
-    
       MReleaseWriteLock(&afs_xdcache);
+
       /* this guy writes chunks, puts back dcache structs, and bumps newDV */
       /* "moredata" just says "there are more dirty chunks yet to come".
        */
       if (j) {
 	static afs_uint32 lp1 = 10000, lp2 = 10000;
 	struct AFSStoreStatus InStatus;
-	struct AFSFetchStatus OutStatus;
-	int doProcessFS = 0;
-	afs_uint32 base, bytes, nchunks;
+	afs_size_t base, bytes;
+	afs_uint32 nchunks;
 	int nomore;
 	unsigned int first;
 	int *shouldwake;
@@ -298,6 +332,7 @@ afs_StoreAllSegments(avc, areq, sync)
 
 	for (bytes = 0, j = 0; !code && j<=high; j++) {
 	  if (dcList[j]) {
+	    ObtainSharedLock(&(dcList[j]->lock), 629);
 	    if (!bytes)
 	      first = j;
 	    bytes += dcList[j]->f.chunkBytes;
@@ -306,9 +341,6 @@ afs_StoreAllSegments(avc, areq, sync)
 		&& dcList[j+1]) {
 		int sbytes = afs_OtherCSize - dcList[j]->f.chunkBytes;
 		bytes += sbytes;
-
-
-
 	    }
 	  }	    
 	  if (bytes && (j==high || !dcList[j+1])) {
@@ -338,17 +370,43 @@ afs_StoreAllSegments(avc, areq, sync)
 		InStatus.Mask |= AFS_FSYNC;
 	    }
 	    tlen = lmin(avc->m.Length, avc->truncPos);
+            afs_Trace4(afs_iclSetp, CM_TRACE_STOREDATA64, 
+		ICL_TYPE_FID, &avc->fid.Fid,
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(base), 
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(bytes), 
+	       	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(tlen));
 
 	    do {
 		stored = 0;
 		tc = afs_Conn(&avc->fid, areq);
 		if (tc) {
+restart:
 #ifdef RX_ENABLE_LOCKS
 		    AFS_GUNLOCK();
 #endif /* RX_ENABLE_LOCKS */
 		    tcall = rx_NewCall(tc->id);
+#ifdef AFS_64BIT_CLIENT
+		    if (!afs_serverHasNo64Bit(tc)) {
+		        code = StartRXAFS_StoreData64(tcall, 
+				(struct AFSFid *) &avc->fid.Fid,
+				&InStatus, base, bytes, tlen);
+		    } else {
+			if (tlen > 0xFFFFFFFF) {
+			    code = EFBIG;
+			} else {
+			    afs_int32 t1, t2, t3;
+			    t1 = base;
+			    t2 = bytes;
+			    t3 = tlen;
+		    	    code = StartRXAFS_StoreData(tcall, 
+					(struct AFSFid *) &avc->fid.Fid,
+					&InStatus, t1, t2, t3);
+			}
+		    }
+#else /* AFS_64BIT_CLIENT */
 		    code = StartRXAFS_StoreData(tcall, (struct AFSFid *) &avc->fid.Fid,
 						&InStatus, base, bytes, tlen);
+#endif /* AFS_64BIT_CLIENT */
 #ifdef RX_ENABLE_LOCKS
 		    AFS_GLOCK();
 #endif /* RX_ENABLE_LOCKS */
@@ -367,6 +425,11 @@ afs_StoreAllSegments(avc, areq, sync)
 			storeallmissing++;
 			continue; /* panic? */
 		    }
+	            afs_Trace4(afs_iclSetp, CM_TRACE_STOREALL2,
+		       	ICL_TYPE_POINTER, avc, 
+			ICL_TYPE_INT32, tdc->f.chunk,
+			ICL_TYPE_INT32, tdc->index,
+			ICL_TYPE_INT32, tdc->f.inode);
   		    shouldwake = 0;
   		    if (nomore) {
  		       if (avc->asynchrony == -1) {
@@ -439,10 +502,16 @@ afs_StoreAllSegments(avc, areq, sync)
 		        }
 		      }
 #else
-		    code = afs_CacheStoreProc(tcall, tfile, tdc->f.chunkBytes, avc, 
-					      shouldwake, &lp1, &lp2);
+		    code = afs_CacheStoreProc(tcall, tfile, tdc->f.chunkBytes, 
+						avc, shouldwake, &lp1, &lp2);
 #endif /* AFS_NOSTATS */
 		    afs_CFileClose(tfile);
+#ifdef AFS_64BIT_CLIENT
+		    if (code == RXGEN_OPCODE) {
+			afs_serverSetNo64Bit(tc);
+			goto restart;
+		    }
+#endif /* AFS_64BIT_CLIENT */
 		    if ((tdc->f.chunkBytes < afs_OtherCSize) && 
 			(i < (nchunks-1))) {
                        int bsent, tlen, tlen1=0, sbytes = afs_OtherCSize - tdc->f.chunkBytes;
@@ -477,6 +546,7 @@ afs_StoreAllSegments(avc, areq, sync)
                      * one which has the writing bit turned on. */
 		}
 		if (!code) {
+		    struct AFSFetchStatus OutStatus;
 		    struct AFSVolSync tsync;
 #ifdef RX_ENABLE_LOCKS
 		    AFS_GUNLOCK();
@@ -487,7 +557,19 @@ afs_StoreAllSegments(avc, areq, sync)
 #endif /* RX_ENABLE_LOCKS */
 		    hadd32(newDV, 1);
 		    XSTATS_END_TIME;
-		    if (!code) doProcessFS = 1; /* Flag to run afs_ProcessFS() later on */		    
+      
+		    /* Now copy out return params */
+		    UpgradeSToWLock(&avc->lock,28);    /* keep out others for a while */
+		    if (!code) {  /* must wait til RPC completes to be sure of this info */
+			afs_ProcessFS(avc, &OutStatus, areq);
+			/* Keep last (max) size of file on server to see if
+			 * we need to call afs_StoreMini to extend the file.
+			 */
+			if (!moredata)
+			    maxStoredLength = OutStatus.Length;
+
+		    }
+		    ConvertWToSLock(&avc->lock);
 		}
 		if (tcall) {
 #ifdef RX_ENABLE_LOCKS
@@ -507,6 +589,10 @@ afs_StoreAllSegments(avc, areq, sync)
 		tdc = dclist[i];
 		if (!code) {
 		    if (afs_indexFlags[tdc->index] & IFDataMod) {
+			/*
+			 * LOCKXXX -- should hold afs_xdcache(W) when
+			 * modifying afs_indexFlags.
+			 */
 			afs_indexFlags[tdc->index] &= ~IFDataMod;
 			afs_stats_cmperf.cacheCurrDirtyChunks--;
 			afs_indexFlags[tdc->index] &= ~IFDirtyPages;
@@ -522,28 +608,24 @@ afs_StoreAllSegments(avc, areq, sync)
                         }
 		    }
 		}
+		UpgradeSToWLock(&tdc->lock, 628);
 		tdc->f.states &= ~DWriting;  /* correct?*/
-		tdc->flags |= DFEntryMod;
-		lockedPutDCache(tdc);
-	    }
-
-	    if (doProcessFS) {
-		/* Now copy out return params */
-		UpgradeSToWLock(&avc->lock,28);    /* keep out others for a while */
-		afs_ProcessFS(avc, &OutStatus, areq);
-		/* Keep last (max) size of file on server to see if
-		 * we need to call afs_StoreMini to extend the file.
-		 */
-		if (!moredata)
-		    maxStoredLength = OutStatus.Length;
-		ConvertWToSLock(&avc->lock);
-		doProcessFS = 0;
+		tdc->dflags |= DFEntryMod;
+		ReleaseWriteLock(&tdc->lock);
+		afs_PutDCache(tdc);
+		/* Mark the entry as released */
+		dclist[i] = NULL;
 	    }
 
 	    if (code) {
-		for (j++; j<=high; j++)
-		    if ( dcList[j] )
-		    	lockedPutDCache(dcList[j]);
+		for (j++; j<=high; j++) {
+		    if ( dcList[j] ) {
+			ReleaseSharedLock(&(dcList[j]->lock));
+		    	afs_PutDCache(dcList[j]);
+			/* Releasing entry */
+			dcList[j] = NULL;
+		    }
+		}
 	    }
 
 	    afs_Trace2(afs_iclSetp, CM_TRACE_STOREALLDCDONE,
@@ -551,6 +633,18 @@ afs_StoreAllSegments(avc, areq, sync)
 	    bytes = 0;
 	  }
 	}
+	
+	/* Release any zero-length dcache entries in our interval
+	 * that we locked but didn't store back above.
+	 */
+	for (j = 0; j<=high; j++) {
+	  tdc = dcList[j];
+	  if (tdc) {
+	    osi_Assert(tdc->f.chunkBytes == 0);
+	    ReleaseSharedLock(&tdc->lock);
+	    afs_PutDCache(tdc);
+	  }
+	}
       } /* if (j) */
 
     minj += NCHUNKSATONCE;
@@ -579,18 +673,52 @@ afs_StoreAllSegments(avc, areq, sync)
    * A lot of this could be integrated into the loop above 
    */
     if (!code) {
-	afs_hyper_t h_unset;
-	hones(h_unset);
-      MObtainWriteLock(&afs_xdcache,285);  /* overkill, but it gets the 
-					* lock in case GetDSlot needs it */
-      for(safety = 0, index = afs_dvhashTbl[hash]; 
-	  index != NULLIDX && safety < afs_cacheFiles+2;) {
+      afs_hyper_t h_unset;
+      hones(h_unset);
 
-	if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
-	  tdc = afs_GetDSlot(index, 0);
+      minj = 0;
+
+      do {
+	moredata = FALSE;
+	memset((char *)dcList, 0, NCHUNKSATONCE * sizeof(struct dcache *));
+
+	/* overkill, but it gets the lock in case GetDSlot needs it */
+	MObtainWriteLock(&afs_xdcache,285);
+
+	for(j = 0, safety = 0, index = afs_dvhashTbl[hash]; 
+	    index != NULLIDX && safety < afs_cacheFiles+2;) {
+
+	    if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
+		tdc = afs_GetDSlot(index, 0);
+		ReleaseReadLock(&tdc->tlock);
+
+		if (!FidCmp(&tdc->f.fid, &avc->fid) && tdc->f.chunk >= minj) {
+		    off = tdc->f.chunk - minj;
+		    if (off < NCHUNKSATONCE) {
+			/* this is the file, and the correct chunk range */
+			if (j >= NCHUNKSATONCE)
+			    osi_Panic("Too many dcache entries in range\n");
+			dcList[j++] = tdc;
+		    } else {
+			moredata = TRUE;
+			afs_PutDCache(tdc);
+			if (j == NCHUNKSATONCE)
+			    break;
+		    }
+		} else {
+		    afs_PutDCache(tdc);
+		}
+	    }
+
+	    index = afs_dvnextTbl[index];
+	}
+	MReleaseWriteLock(&afs_xdcache);
+
+	for (i=0; i<j; i++) {
+	    /* Iterate over the dcache entries we collected above */
+	    tdc = dcList[i];
+	    ObtainSharedLock(&tdc->lock, 677);
 
-	  if (!FidCmp(&tdc->f.fid, &avc->fid)) {
-	    /* this is the file */
 	    /* was code here to clear IFDataMod, but it should only be done
 	     * in storedcache and storealldcache.
 	     */
@@ -602,21 +730,27 @@ afs_StoreAllSegments(avc, areq, sync)
 	     * for a chunk that we didn't store this time
 	     */
 	     /* Don't update the version number if it's not yet set. */
-	    if (code == 0 && (!hsame(tdc->f.versionNo, h_unset))
-	        && (hcmp(tdc->f.versionNo, oldDV) >= 0)) {
-	      if ((!(afs_dvhack || foreign) && hsame(avc->m.DataVersion, newDV))
-		  || ((afs_dvhack || foreign) && (origCBs == afs_allCBs)) ) {
-		  /* no error, this is the DV */
-	        hset(tdc->f.versionNo, avc->m.DataVersion);
-	        tdc->flags |= DFEntryMod;
-	      }
+	    if (!hsame(tdc->f.versionNo, h_unset) &&
+		hcmp(tdc->f.versionNo, oldDV) >= 0) {
+
+	    	if ((!(afs_dvhack || foreign) && hsame(avc->m.DataVersion, newDV))
+		    || ((afs_dvhack || foreign) && (origCBs == afs_allCBs)) ) {
+		    /* no error, this is the DV */
+
+		    UpgradeSToWLock(&tdc->lock, 678);
+		    hset(tdc->f.versionNo, avc->m.DataVersion);
+		    tdc->dflags |= DFEntryMod;
+		    ConvertWToSLock(&tdc->lock);
+		}
 	    }
-	  }
-	  lockedPutDCache(tdc);
+
+	    ReleaseSharedLock(&tdc->lock);
+	    afs_PutDCache(tdc);
 	}
-	index = afs_dvnextTbl[index];
-      }
-      MReleaseWriteLock(&afs_xdcache);
+
+	minj += NCHUNKSATONCE;
+
+      } while (moredata);
     }
 
     if (code) {
@@ -627,7 +761,7 @@ afs_StoreAllSegments(avc, areq, sync)
 	 * fileserver.
 	 */
 	if (areq->permWriteError || (avc->states & (CCore1 | CCore))) {
-	    afs_InvalidateAllSegments(avc, 1);
+	    afs_InvalidateAllSegments(avc);
 	}
     }
     afs_Trace3(afs_iclSetp, CM_TRACE_STOREALLDONE, ICL_TYPE_POINTER, avc,
@@ -674,33 +808,32 @@ afs_StoreAllSegments(avc, areq, sync)
  *
  * Parameters:
  *	avc      : Pointer to vcache entry.
- *	asetLock : If true, we are to set the afs_xdcache lock; otherwise,
- *		   the caller has already done it.
  *
  * Environment:
  *	For example, called after an error has been detected.  Called
- *	with avc write-locked.
+ *	with avc write-locked, and afs_xdcache unheld.
  */
    
-afs_InvalidateAllSegments(avc, asetLock)
+afs_InvalidateAllSegments(avc)
     struct vcache *avc;
-    int asetLock;
 
 { /*afs_InvalidateAllSegments*/
 
     struct dcache *tdc;
     afs_int32 hash;
     afs_int32 index;
+    struct dcache **dcList;
+    int i, dcListMax, dcListCount;
 
     AFS_STATCNT(afs_InvalidateAllSegments);
     afs_Trace2(afs_iclSetp, CM_TRACE_INVALL, ICL_TYPE_POINTER, avc,
-	       ICL_TYPE_INT32, avc->m.Length);
+	       ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(avc->m.Length));
     hash = DVHash(&avc->fid);
     avc->truncPos = AFS_NOTRUNC;  /* don't truncate later */
     avc->states &= ~CExtendedFile; /* not any more */
     ObtainWriteLock(&afs_xcbhash, 459);
     afs_DequeueCallback(avc);
-    avc->states &= ~(CStatd|CDirty);	  /* mark status information as bad, too */
+    avc->states &= ~(CStatd|CDirty); /* mark status information as bad, too */
     ReleaseWriteLock(&afs_xcbhash);
     if (avc->fid.Fid.Vnode & 1 || (vType(avc) == VDIR))
 	osi_dnlc_purgedp(avc);
@@ -713,28 +846,60 @@ afs_InvalidateAllSegments(avc, asetLock)
      * Block out others from screwing with this table; is a read lock
      * sufficient?
      */
-    if (asetLock) MObtainWriteLock(&afs_xdcache,286);
+    MObtainWriteLock(&afs_xdcache,286);
+    dcListMax = 0;
+
     for(index = afs_dvhashTbl[hash]; index != NULLIDX;) {
-      if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
-	tdc = afs_GetDSlot(index, 0);
-	if (!FidCmp(&tdc->f.fid, &avc->fid)) {
-	    /* same file? we'll zap it */
-	    if (afs_indexFlags[index] & IFDataMod) {
-		afs_stats_cmperf.cacheCurrDirtyChunks--;
-		/* don't write it back */
-		afs_indexFlags[index] &= ~IFDataMod;
-	      }
-	    afs_indexFlags[index] &= ~IFAnyPages;
-	    ZapDCE(tdc);
-	    if (vType(avc) == VDIR) {
-		DZap(&tdc->f.inode);
+	if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
+	    tdc = afs_GetDSlot(index, 0);
+	    ReleaseReadLock(&tdc->tlock);
+	    if (!FidCmp(&tdc->f.fid, &avc->fid))
+		dcListMax++;
+	    afs_PutDCache(tdc);
+	}
+	index = afs_dvnextTbl[index];
+    }
+
+    dcList = osi_Alloc(dcListMax * sizeof(struct dcache *));
+    dcListCount = 0;
+
+    for(index = afs_dvhashTbl[hash]; index != NULLIDX;) {
+	if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
+	    tdc = afs_GetDSlot(index, 0);
+	    ReleaseReadLock(&tdc->tlock);
+	    if (!FidCmp(&tdc->f.fid, &avc->fid)) {
+		/* same file? we'll zap it */
+		if (afs_indexFlags[index] & IFDataMod) {
+		    afs_stats_cmperf.cacheCurrDirtyChunks--;
+		    /* don't write it back */
+		    afs_indexFlags[index] &= ~IFDataMod;
+		}
+		afs_indexFlags[index] &= ~IFAnyPages;
+		if (dcListCount < dcListMax)
+		    dcList[dcListCount++] = tdc;
+		else
+		    afs_PutDCache(tdc);
+	    } else {
+		afs_PutDCache(tdc);
 	    }
 	}
-	lockedPutDCache(tdc);
-      }
-      index = afs_dvnextTbl[index];
+	index = afs_dvnextTbl[index];
     }
-    if (asetLock) MReleaseWriteLock(&afs_xdcache);
+    MReleaseWriteLock(&afs_xdcache);
+
+    for (i=0; i<dcListCount; i++) {
+	tdc = dcList[i];
+
+	ObtainWriteLock(&tdc->lock, 679);
+	ZapDCE(tdc);
+	if (vType(avc) == VDIR)
+	    DZap(&tdc->f.inode);
+	ReleaseWriteLock(&tdc->lock);
+	afs_PutDCache(tdc);
+    }
+
+    osi_Free(dcList, dcListMax * sizeof(struct dcache *));
+
     return 0;
 
 } /*afs_InvalidateAllSegments*/
@@ -756,7 +921,7 @@ afs_InvalidateAllSegments(avc, asetLock)
  *	held.
  */
 afs_TruncateAllSegments(avc, alen, areq, acred)
-    afs_int32 alen;
+    afs_size_t alen;
     register struct vcache *avc;
     struct vrequest *areq;
     struct AFS_UCRED *acred;
@@ -767,8 +932,14 @@ afs_TruncateAllSegments(avc, alen, areq, acred)
     register afs_int32 index;
     afs_int32 newSize;
 
+    int dcCount, dcPos;
+    struct dcache **tdcArray;
+
     AFS_STATCNT(afs_TruncateAllSegments);
     avc->m.Date = osi_Time();
+    afs_Trace3(afs_iclSetp, CM_TRACE_TRUNCALL, ICL_TYPE_POINTER, avc,
+		   	ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(avc->m.Length), 
+			ICL_TYPE_OFFSET, ICL_HANDLE_OFFSET(alen));
     if (alen >= avc->m.Length) {
 	/*
 	 * Special speedup since Sun's vm extends the file this way;
@@ -781,14 +952,9 @@ afs_TruncateAllSegments(avc, alen, areq, acred)
 	 */
 	avc->states |= CExtendedFile;
 	avc->m.Length = alen;
-	afs_Trace3(afs_iclSetp, CM_TRACE_TRUNCALL1, ICL_TYPE_POINTER, avc,
-		   ICL_TYPE_INT32, avc->m.Length, ICL_TYPE_INT32, alen);
 	return 0;
     }
 
-    afs_Trace3(afs_iclSetp, CM_TRACE_TRUNCALL2, ICL_TYPE_POINTER, avc,
-	       ICL_TYPE_INT32, avc->m.Length, ICL_TYPE_INT32, alen);
-
 #if	(defined(AFS_SUN5_ENV))
 
     /* Zero unused portion of last page */
@@ -815,27 +981,72 @@ afs_TruncateAllSegments(avc, alen, areq, acred)
     
     if (alen < avc->truncPos) avc->truncPos = alen;
     code = DVHash(&avc->fid);
+
     /* block out others from screwing with this table */
     MObtainWriteLock(&afs_xdcache,287);
+
+    dcCount = 0;
     for(index = afs_dvhashTbl[code]; index != NULLIDX;) {
       if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
 	tdc = afs_GetDSlot(index, 0);
-	if (!FidCmp(&tdc->f.fid, &avc->fid)) {
-	    /* same file, and modified, we'll store it back */
-	    newSize = alen - AFS_CHUNKTOBASE(tdc->f.chunk);
-	    if (newSize < 0) newSize = 0;
-	    if (newSize < tdc->f.chunkBytes) {
-		register struct osi_file *tfile;
-		tfile = afs_CFileOpen(tdc->f.inode);
-		afs_CFileTruncate(tfile, newSize);
-		afs_CFileClose(tfile);
-		afs_AdjustSize(tdc, newSize);
-	    }
-	}
-	lockedPutDCache(tdc);
+	ReleaseReadLock(&tdc->tlock);
+	if (!FidCmp(&tdc->f.fid, &avc->fid))
+	    dcCount++;
+	afs_PutDCache(tdc);
       }
       index = afs_dvnextTbl[index];
     }
+
+    /* Now allocate space where we can save those dcache entries, and
+     * do a second pass over them..  Since we're holding xdcache, it
+     * shouldn't be changing.
+     */
+    tdcArray = osi_Alloc(dcCount * sizeof(struct dcache *));
+    dcPos = 0;
+
+    for(index = afs_dvhashTbl[code]; index != NULLIDX;) {
+      if (afs_indexUnique[index] == avc->fid.Fid.Unique) {
+	tdc = afs_GetDSlot(index, 0);
+	ReleaseReadLock(&tdc->tlock);
+	if (!FidCmp(&tdc->f.fid, &avc->fid)) {
+	    /* same file, and modified, we'll store it back */
+	    if (dcPos < dcCount) {
+		tdcArray[dcPos++] = tdc;
+	    } else {
+		afs_PutDCache(tdc);
+	    }
+	} else {
+	    afs_PutDCache(tdc);
+	}
+      }
+      index = afs_dvnextTbl[index];
+    }
+
+    MReleaseWriteLock(&afs_xdcache);
+
+    /* Now we loop over the array of dcache entries and truncate them */
+    for (index = 0; index < dcPos; index++) {
+	struct osi_file *tfile;
+
+	tdc = tdcArray[index];
+
+	newSize = alen - AFS_CHUNKTOBASE(tdc->f.chunk);
+	if (newSize < 0) newSize = 0;
+	ObtainSharedLock(&tdc->lock, 672);
+	if (newSize < tdc->f.chunkBytes) {
+	    UpgradeSToWLock(&tdc->lock, 673);
+	    tfile = afs_CFileOpen(tdc->f.inode);
+	    afs_CFileTruncate(tfile, newSize);
+	    afs_CFileClose(tfile);
+	    afs_AdjustSize(tdc, newSize);
+	    ConvertWToSLock(&tdc->lock);
+	}
+	ReleaseSharedLock(&tdc->lock);
+	afs_PutDCache(tdc);
+    }
+
+    osi_Free(tdcArray, dcCount * sizeof(struct dcache *));
+
 #if	(defined(AFS_SUN5_ENV))
     ObtainWriteLock(&avc->vlock, 547);
     if (--avc->activeV == 0 && (avc->vstates & VRevokeWait)) {
@@ -844,7 +1055,6 @@ afs_TruncateAllSegments(avc, alen, areq, acred)
     }
     ReleaseWriteLock(&avc->vlock);
 #endif
-    MReleaseWriteLock(&afs_xdcache);
     return 0;
 
 } /*afs_TruncateAllSegments*/
